/*
 * record 1 speech segment from microphone input
 * and save to a file
 *
 */

#include <sent/stddefs.h>
#include <sent/speech.h>
#include <sent/adin.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


#if defined(USE_MIC) || defined(USE_NETAUDIO)

static SP16 speech[MAXSPEECHLEN]; /* buffer to store speech data */
static int speechlen;		/* data length of above */
static boolean strip_zero_sample = TRUE; /* strip off zero samples */
static boolean stout = FALSE;	/* standard output for filename "-" */
#ifdef HAVE_PTHREAD
/* use thread ad-in */
#include <pthread.h>
static boolean transp = FALSE;	/* TRUE when thread writing to speech[] */
static pthread_t adin_id;	/* thread ID for thread ad-in */
static ADIN_THREAD_P targ;	/* argument to pass to ad-in thread */
#endif /* HAVE_PTHREAD */

#ifndef HAVE_PTHREAD
/* only store to buffer */
static boolean
adin_callback(SP16 *now, int len)
{
  if (speechlen + len > MAXSPEECHLEN) {
    fprintf(stderr, "Error: too long input (> %d samples)\n", MAXSPEECHLEN);
    return(FALSE);
  }
  memcpy(&(speech[speechlen]), now, len * sizeof(SP16));
  if (stout) {
    wrsamp(1, &(speech[speechlen]), len);
  }
  speechlen += len;
  /* output progress bar in dots */
  fprintf(stderr, ".");
  return(TRUE);
}
#endif

void
usage()
{
  fprintf(stderr, "adinrec --- record one speech to file (16kHz,16bit,mono)\n");
  fprintf(stderr, "options:");
#ifdef USE_MIC
  fprintf(stderr, " MIC");
#endif
#ifdef USE_NETAUDIO
  fprintf(stderr, " NETAUDIO");
#endif
#ifdef HAVE_PTHREAD
  fprintf(stderr, " PTHREAD");
#endif
  fprintf(stderr, "\n");

  fprintf(stderr, "Usage: adinrec [-lv threshold(def:3000)]");
#ifdef USE_NETAUDIO
  fprintf(stderr, " [-input netaudio -NA host:unit]");
#endif
  fprintf(stderr, " [-nostrip]");
  fprintf(stderr, " filename\n");
  exit(1);
}
  

int
main(int argc, char *argv[])
{
  size_t size;
  char *filename;
  int thres = 3000;	/* input trigger level threshold */
  int i;
  int src=1;			/* 1: mic 2: netaudio */
#ifdef USE_NETAUDIO
  char *netaudio_devname = NULL;
#endif
  
  /* option parsing */
  if (argc <= 1) usage();
  for (i=1;i<argc;i++) {
    if (!strcmp(argv[i], "-lv")) {
      i++;
      if (i >= argc) usage();
      thres = atoi(argv[i]);
    } else if (!strcmp(argv[i], "-input")) {
      i++;
      if (i >= argc) usage();
      if (!strcmp(argv[i], "mic")) {
#ifdef USE_MIC
	src = 1;
#else
	fprintf(stderr,"Error: mic input not available\n");
	usage();
#endif
      } else if (!strcmp(argv[i], "netaudio")) {
#ifdef USE_NETAUDIO
	src = 2;
#else
	fprintf(stderr,"Error: netaudio input not available\n");
	usage();
#endif
      }
    } else if (!strcmp(argv[i], "-NA")) {
#ifdef USE_NETAUDIO
      i++;
      if (i >= argc) usage();
      netaudio_devname = argv[i];
#else
      fprintf(stderr,"Error: netaudio input not available\n");
      usage();
#endif
    } else if (!strcmp(argv[i], "-nostrip")) {
      strip_zero_sample = FALSE;
    } else {
      if (!strcmp(argv[i], "-")) {
	stout = TRUE;
      } else {
	filename = argv[i];
      }
    }
  }
  
  /* initialize port */
#ifdef USE_MIC
  if (src == 1) {
    /* select microphone device */
    adin_setup_adfunc(adin_mic_start, adin_mic_stop, adin_mic_read, thres);
    if (adin_mic_standby() == FALSE) {
      fprintf(stderr,"*** mic init failed! ***\n");
      return(1);
    }
  }
#endif
#ifdef USE_NETAUDIO
  if (src == 2) {
    /* select microphone device */
    adin_setup_adfunc(adin_netaudio_start, adin_netaudio_stop, adin_netaudio_read, thres);
    if (adin_netaudio_standby(netaudio_devname) == FALSE) {
      fprintf(stderr,"*** NetAudio init failed! ***\n");
      return(1);
    }
  }
#endif
  if (strip_zero_sample) {
    fprintf(stderr, "Zero frames stripping: on\n");
  } else {
    fprintf(stderr, "Zero frames stripping: off\n");
  }
#ifdef HAVE_PTHREAD
  targ.speech = speech;
  targ.len = &speechlen;
  targ.maxlen = MAXSPEECHLEN;
  targ.stflag = strip_zero_sample;
  targ.tflag = &transp;
  pthread_create(&adin_id, NULL, (void *)adin_cut_thread, (void *)&targ);
  pthread_detach(adin_id); /* not join, run forever */
  fprintf(stderr,"AD-in thread created\n");
#endif
  fprintf(stderr,"init ok, start recording...\n");
  
  /* get 1 speech segment to buffer (not realtime) */
#ifdef HAVE_PTHREAD
  {
    int oldlen, tmplen;
    speechlen = 0;
    /* tell adin thread to begin recording */
    transp = TRUE;
    /* wait for adin thread to end */
    while((tmplen = speechlen) != oldlen || transp == TRUE) {
      if (oldlen != tmplen) { /* thread is recording now */
	if (stout) {
	  wrsamp(1, &(speech[oldlen]), tmplen - oldlen);
	}
	oldlen = tmplen;
	fprintf(stderr, ".");
      }
      usleep(100000);		/* wait = 0.1sec */
    }
  }
#else  /* ~HAVE_PTHREAD */
  speechlen = 0;
  adin_cut(adin_callback, strip_zero_sample);
#endif
  size = sizeof(SP16) * speechlen;
  if (stout) {
    fprintf(stderr, "%d samples (%d bytes) recorded to stdout\n", speechlen, size);
  } else {
    fprintf(stderr, "%d samples (%d bytes) recorded on memory\n", speechlen, size);
    /* write to file */
    fprintf(stderr, "writing to %s..\n",filename);
    {
      int fd;
      size_t count;
      if (access(filename, F_OK) == 0) {
	if (access(filename, W_OK) == 0) {
	  fprintf(stderr, "overwriting...\n");
	} else {
	  perror("adin");
	  return(1);
	}
      } 
      if ((fd = creat(filename, 0644)) == -1) {perror("adin"); return 1;}
      size = sizeof(SP16) * speechlen;
      count = wrsamp(fd, speech, speechlen);
      if (count < 0) {perror("adin: cannot write"); return 1;}
      if (count < size) {fprintf(stderr, "adin: cannot fully write %d bytes\n", size); return 1;}
      if (close(fd) < 0) {perror("adin"); return 1;}
      fprintf(stderr, "done\n");
    }
  }
  
  return 0;
}

#else  /* ~USE_MIC */

int
main(int argc, char *argv[])
{
  fprintf(stderr,"neither microphone nor NetAudio is supported on this OS\n");
  return(1);
}  

#endif /* ~USE_MIC */
