/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* realtime-1stpass.c --- on-the-fly 1st pass processing with realtime CMN */
/* do speech input, make mfcc (attach), viterbi in one loop */
/* the main RealTimePipeLine can be called incrementally as a */
/* callback from adin_cut() */

#include <julius.h>

/* MFCC parameter */
static HTK_Param *param = NULL;	/* hold MFCC param */
static Value para;		/* parameters for Wav2MFCC */
static float *bf;		/* work space for FFT */

static boolean need_attach = FALSE; /* whether param need selection */
static int maxframelen;		/* maximum frame length from MAXSPEECHLEN */


/**** progress pointers ****/
/*
 * [SPEECH DATA]
 *    0               nowlen
 *    |---------------|
 * ooo.....
 *   o........
 *      ........
 *        ........
 *
 *
 * [PARAMETER VECTOR]
 *     O o o o o o f_raw
 *     O o o o o o
 *     O . . . . .                        (MFCC+RAW_E 13)
 *     O o o o o o                        
 *     O o o o |-|para.delWin
 *     O o o o                            (MFCC+RAW_E+delta MFCC + delta E 26)
 *     O o o o
 *       o o o
 *     |---------|f_delay = para.delWin * 2 + 1
 *     
 *     f ... (MFCC + delta MFCC + delta E 25) + realtime-CMN
 */   

static int f_raw;
static int f_delay;
static int f;
static SP16 *window;
static int windowlen;
static int windownum;		/* current left samples in a window */

/* set MFCC parameter selection for IPA model */
static void
init_para()
{
  para.smp_period = DEF_SMPPERIOD;
  para.framesize  = DEF_FRAMESIZE;
  para.frameshift = DEF_FRAMESHIFT;
  para.preEmph    = DEF_PREENPH;
  para.mfcc_dim   = DEF_MFCCDIM;
  para.lifter     = DEF_CEPLIF;
  para.delWin     = DEF_DELWIN;
  para.silFloor   = DEF_SILFLOOR;
  /* para.escale     = DEF_ESCALE; */
  para.escale     = 1.0;
  para.fbank_num  = DEF_FBANK;
  /* para.cmn        = FALSE;*/
  para.cmn        = TRUE;
  para.enormal    = FALSE;
  para.raw_e      = FALSE;
  para.vec_num    = (para.mfcc_dim + 1) * 2;

}

/* prepare parameter vector holder */
static void
init_param()
{
  param = new_param();
  param->header.samptype = F_MFCC | F_ENERGY | F_DELTA | F_CEPNORM;
  param->header.wshift = 100000; /* 16kHz */
  param->header.sampsize = para.vec_num * sizeof(VECT); /* not compressed */
  param->veclen = para.vec_num;
  /* assign max (safe with free_param)*/
  param->parvec = (VECT **)mymalloc(sizeof(VECT *) * maxframelen);
  /* param->parvec */
  /* param->header.samplenum */
  /* param->samplenum */
}


/* initialize (once on startup) */
void
RealTimeInit()
{
  init_para();
  WMP_init(para, &bf);
  f_delay = para.delWin * 2 + 1;
  /* pre-fetch outprob cache for maximum length */
  maxframelen = MAXSPEECHLEN / para.frameshift;
  if (hmminfo.is_tied_mixture) {
    mixture_cache_prepare(maxframelen);
  }
  outprob_cache_prepare(maxframelen);
  windowlen = para.framesize + 1;
  window = mymalloc(sizeof(SP16) * windowlen);
}

/* prepare (on start of every loop) */
void
RealTimePipeLinePrepare()
{
  /* beginning of speech */
  init_param();
  outprob_cache_clear(maxframelen); /* no re-allocate */
  if (!check_param_coherence(&hmminfo,param)) {
    need_attach = TRUE;
    if (select_param_vmark(param, hmminfo.opt.param_type) == FALSE) {
      j_printerr("cannot attach Param!\n");
      exit(1);
    }
    param->header.samptype = hmminfo.opt.param_type;
  }
  f_raw = 0;
  f = 0;
  windownum = 0;
}

/* main rountine */
boolean
RealTimePipeLine(SP16 *Speech, int nowlen)
{
  int i;
  int now;

  /* bf[1-bfnum] is left from previous call */
  
  now = 0;

  while (now < nowlen) {
    /* fill as many as possible */
    for(i = min(windowlen - windownum, nowlen - now); i > 0 ; i--)
      window[windownum++] = (float) Speech[now++];
    if (windownum < windowlen) break; /* shortage or last */

    /* needed samples -> bf[] */
    for (i=0; i < windowlen; i++) {
      bf[i+1] = (float) window[i];
    }

    /* calc wave -> MFCC_E_Z */
    /* needs conversion here if integerized */
    param->parvec[f_raw] = (VECT *)mymalloc(sizeof(VECT) * para.vec_num);
    WMP_calc(param->parvec[f_raw], bf, para);
    
    if (f_raw >= para.delWin) {
      /* delay calc MFCC_E -> MFCC_E_D_Z */
      /* needs conversion here if integerized */
      WMP_Delta(param->parvec, f_raw-para.delWin, f_raw+1, para);
    }
    if (f_raw >= f_delay) {
      /* more-delayed calc MFCC_E_D_Z -> MFCC_E_D_N_Z with CMN */
      f = f_raw - f_delay;
      if(para.cmn) CMN_realtime(param->parvec[f], para.mfcc_dim);
      if (need_attach) {
	param->veclen = exec_exclude_one_vector(param->parvec[f], para.vec_num);
	param->header.sampsize = param->veclen * sizeof(VECT);
      }
	
      /* proceed beam for 1 frame */
      if (f == 0) {
	get_back_trellis_init(param, &backtrellis);
      } else {
	if (get_back_trellis_proceed(f, param, &backtrellis) == FALSE) {
	  return FALSE;
	}
      }
    }
    
    memmove(window, &(window[para.frameshift]), sizeof(SP16) * (windowlen - para.frameshift));
    windownum -= para.frameshift;
    f_raw++;
  }
  
  return(TRUE);
}

/* end proc and return obtained param */
HTK_Param *
RealTimeParam(LOGPROB *backmax)
{
  /* flush rest samples */
  for (f = f_raw - para.delWin; f < f_raw; f++){
    WMP_Delta(param->parvec, f, f_raw, para);
  }
  for (f = f_raw - f_delay; f < f_raw; f++) {
    if(para.cmn) CMN_realtime(param->parvec[f], para.mfcc_dim);
    if (need_attach) {
      param->veclen = exec_exclude_one_vector(param->parvec[f], para.vec_num);
      param->header.sampsize = param->veclen * sizeof(VECT);
    }
    get_back_trellis_proceed(f, param, &backtrellis);
  }
  /* end procs */
  param->header.samplenum = f_raw;
  param->samplenum = f_raw;
  get_back_trellis_end(param, &backtrellis, backmax);

  /* update CMN vector for next speech */
  if(para.cmn) CMN_realtime_update();

  /* return obtained parameter for 2nd pass */
  return(param);
}
