/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* word_align.c --- perform viterbi alignment for given words or phonemes */

#include <julius.h>

/* build sentence HMM from word sequence */

static HMM_Logical **
make_phseq(WORD_ID *wseq, short num, int *num_ret, int **end_ret, boolean per_phoneme)
{
  HMM_Logical **ph;		/* phoneme sequence */
  int phnum;			/* num of above */
  WORD_ID tmpw, w;
  int i, pn, st, endn;
  HMM_Logical *tmpp;

  /* make ph[] from wseq[] */
  /* 1. calc total phone num and malloc */
  phnum = 0;
  for (w=0;w<num;w++) phnum += winfo.wlen[wseq[w]];
  ph = (HMM_Logical **)mymalloc(sizeof(HMM_Logical *) * phnum);
  /* 2. make phoneme sequence */
  st = 0;
  pn = 0;
  endn = 0;
  for (w=0;w<num;w++) {
    tmpw = wseq[w];
    for (i=0;i<winfo.wlen[tmpw];i++) {
      tmpp = winfo.wseq[tmpw][i];
      /* handle cross-word context dependency */
#ifdef CCD_ON_2PASS
      if (ccd_flag) {
	if (w > 0 && i == 0) {	/* word head */
	  tmpp = get_left_context_HMM(tmpp, ph[pn-1], &hmminfo);
	}
	if (w < num-1 && i == winfo.wlen[tmpw] - 1) { /* word tail */
	  tmpp = get_right_context_HMM(tmpp, winfo.wseq[wseq[w+1]][0], &hmminfo);
	}
      }
#endif /* CCD_ON_2PASS */
      ph[pn++] = tmpp;
      st += tmpp->def->state_num - 2;
      if (per_phoneme) (*end_ret)[endn++] = st - 1;
    }
    if (!per_phoneme) (*end_ret)[endn++] = st - 1;
  }
  *num_ret = phnum;
  return ph;
}

/* build sentence HMM, call viterbi_segment() and output result */
/* NOTE: words are in reverse order */
static void
do_align(WORD_ID *words, short wnum, HTK_Param *param, boolean per_phoneme)
{
  HMM_Logical **phones;		/* phoneme sequence */
  int phonenum;			/* num of above */
  HMM *shmm;			/* sentence HMM */
  int *end_state;		/* state number of word ends */
  int *end_frame;		/* segmented last frame of words */
  LOGPROB *end_score;		/* normalized score of each words */
  LOGPROB allscore;		/* total score of this word sequence */
  WORD_ID w;
  int end_num;

  if (per_phoneme) {
    j_printf("=== phoneme alignment begin ===\n");
  } else {
    j_printf("=== word alignment begin ===\n");
  }

  /* initialize result buffers */
  if (per_phoneme) {
    end_num = 0;
    for(w=0;w<wnum;w++) end_num += winfo.wlen[words[w]];
  } else {
    end_num = wnum;
  }
  end_state = (int *)mymalloc(sizeof(int) * end_num);
  end_frame = (int *)mymalloc(sizeof(int) * end_num);
  end_score = (LOGPROB *)mymalloc(sizeof(LOGPROB) * end_num);

  /* make phoneme sequence word sequence */
  phones = make_phseq(words, wnum, &phonenum, &end_state, per_phoneme);
  /* build the sentence HMMs */
  shmm = new_make_word_hmm(&hmminfo, phones, phonenum);

  /* call viterbi segmentation function */
  allscore = viterbi_segment(shmm, param, end_state, end_num, end_frame, end_score, outprob);

  /* print result */
  {
    int i,p,n;
    j_printf("id: from  to    n_score    applied HMMs (logical[physical])\n");
    j_printf("------------------------------------------------------------\n");
    n = 0;
    for (i=0;i<end_num;i++) {
      j_printf("%2d: %4d %4d  %f ", i, (i == 0) ? 0 : end_frame[i-1]+1, end_frame[i], end_score[i]);
      if (per_phoneme) {
	if (strmatch(phones[n]->name, phones[n]->def->name)) {
	  j_printf(" %s", phones[n]->name);
	} else {
	  j_printf(" %s[%s]", phones[n]->name, phones[n]->def->name);
	}
	n++;
      } else {
	for(p=0;p<winfo.wlen[words[i]];p++) {
	  if (strmatch(phones[n]->name, phones[n]->def->name)) {
	    j_printf(" %s", phones[n]->name);
	  } else {
	    j_printf(" %s[%s]", phones[n]->name, phones[n]->def->name);
	  }
	  n++;
	}
      }
      j_printf("\n");
    }
  }
  j_printf("re-computed AM score: %f\n", allscore);

  free(phones);
  free(end_score);
  free(end_frame);
  free(end_state);

  if (per_phoneme) {
    j_printf("=== phoneme alignment end ===\n");
  } else {
    j_printf("=== word alignment end ===\n");
  }
  
}

void
word_align(WORD_ID *words, short wnum, HTK_Param *param)
{
  do_align(words, wnum, param, FALSE);
}
void
word_rev_align(WORD_ID *revwords, short wnum, HTK_Param *param)
{
  WORD_ID *words;		/* word sequence (true order) */
  int w;
  words = (WORD_ID *)mymalloc(sizeof(WORD_ID) * wnum);
  for (w=0;w<wnum;w++) words[w] = revwords[wnum-w-1];
  do_align(words, wnum, param, FALSE);
  free(words);
}
void
phoneme_align(WORD_ID *words, short wnum, HTK_Param *param)
{
  do_align(words, wnum, param, TRUE);
}
void
phoneme_rev_align(WORD_ID *revwords, short wnum, HTK_Param *param)
{
  WORD_ID *words;		/* word sequence (true order) */
  int w;
  words = (WORD_ID *)mymalloc(sizeof(WORD_ID) * wnum);
  for (w=0;w<wnum;w++) words[w] = revwords[wnum-w-1];
  do_align(words, wnum, param, TRUE);
  free(words);
}
