/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* adin-cut.c --- get audio segment with zerocross & level cutting */

#include <sent/stddefs.h>
#include <sent/speech.h>

#define NOISE_ZEROCROSS 40	/* threshold of zerocross num in a period */

static int c_length  = 5000;	/* samples to hold */
static int c_offset  = 0;	/* data DC offset */
static int wstep   = 1000;	/* data window size */
static int nc_max  = 12;	/* input end if noise period repeat this times */
static int thres;		/* level threshold */

/* pointers to device-dependent API functions */
static boolean (*ad_start)();
static boolean (*ad_stop)();
static int (*ad_read)(SP16 *, int);


/* register device-dependent API functions */
void
adin_setup_adfunc(
	 boolean (*cad_start)(),
	 boolean (*cad_stop)(),
	 int (*cad_read)(SP16 *, int),
         int cthres)
{
  ad_start = cad_start;
  ad_stop = cad_stop;
  ad_read = cad_read;
  thres = cthres;
}

/* get audio segment from input source with zero-cross cutting' */
/* for each incoming segment, calls the callback function */
boolean
adin_cut(boolean (*ad_process)(SP16 *, int), boolean strip_flag)
{
  SP16 *buffer;
  int bp, bpmax;
  int len_limit, realtime_factor;
  int i, imax, len, stlen;
  int cnt;
  
  /* variables for zero-cross routines*/
  int zc;
  boolean is_valid_data = FALSE;
  int nc = 0;

  /* there are 3 buffers:
   *   temporary storage queue: buffer[]
   *   cycle buffer for zero-cross counting: (in zc_e)
   *
   * Each samples are first read to buffer[], then passed to count_zc_e()
   * to find trigger.  Samples between trigger and end of speech are 
   * passwd to (*ad_process) with pointer to the first byte and its length.
   *
   * All procedure is done in pipe-line style.
   *
   */
  /* prepare temporary buffer (queue) */
  bpmax = MAXSPEECHLEN;
  buffer = (SP16 *)mymalloc(sizeof(SP16) * bpmax);
  bp = 0;

  /** initialize params. of zero cross method and etc **/
  init_count_zc_e(thres, c_length, c_offset);

  /* prepare audio device */
  if ((*ad_start)() == FALSE) {
    j_printerr("Error: failed to start recording\n");
    free(buffer);
    return(FALSE);
  }

  realtime_factor = 1;		/* assume RTx 1 */
  len_limit = 16000;

  /* ---------------------------------------------- */
  /* main loop */
  for (;;) {

    /* Read samples as 16bit shorts (SP16) */
    /* bp: pointer to samples left in queue buffer */
    /* all samples in device are read at offset [bp] -> [0..len-1] */
    cnt = (*ad_read)(&(buffer[bp]), bpmax - bp);
    if (cnt < 0) {
      j_printerr("sample read error\n");
      exit(-1);
    }
    len = bp + cnt;
    /* j_printf("- read %d (%d-%d)\n",cnt,bp,len); */

    /* strip off invalid samples */
    if (strip_flag) {
      stlen = strip_zero(&(buffer[bp]), cnt);
      if (stlen != cnt) {
	cnt = stlen;
	len = bp + cnt;
      }
    }

    /* adjust amounts of process per a loop */
    if (cnt > 16000) {		/* more than 1 sec. samples in device */
      /* queueing is getting delayed when process is slower than realtime. */
      /* now increases limits to the amount of processing samples per a loop */
      /* to avoid buffer overrun in AD-in device */
      realtime_factor++;
      len_limit = cnt / realtime_factor;
    }
    imax = (len < len_limit) ? len : len_limit;

    /* proceed for each 'wstep' steps */
    i = 0;
    while (i + wstep <= imax) {
      /* count zero-cross and flip switches */
      /* swap samples in buffer to that of (length) step ago */
      zc = count_zc_e(&(buffer[i]), wstep);
      if (zc > NOISE_ZEROCROSS) {
	nc = 0;                   /* reset nc counter */
	if (is_valid_data == FALSE) {
	  is_valid_data = TRUE;   /* start to record */
	}
      } else if (is_valid_data == TRUE) {
	/* noise after input */
	nc++;
	if (nc >= nc_max) {       /* end of loop */
	  goto break_input;
	}
      }
      if (is_valid_data == TRUE) {
	/* store data */
	/*memcpy(&(speech[total]), &(buffer[i]), wstep * sizeof(SP16));*/
	if ( ad_process != NULL ) {
	  /* call external function */
	  if ((*ad_process)(&(buffer[i]),  wstep) == FALSE) goto break_input;
	}
      }
      i += wstep;
    }
    
    /* purge processed samples and update queue */
    if (i > 0) {
      memmove(buffer, &(buffer[i]), (len - i) * sizeof(SP16));
    }
    bp = len - i;
  }

break_input:
  /* pause transfer */
  if ((*ad_stop)() == FALSE) {
    j_printerr("Error: failed to stop recording\n");
    free(buffer);
    end_count_zc_e();
    return(FALSE);
  }
  free(buffer);
  end_count_zc_e();
  return(TRUE);
}



#ifdef HAVE_PTHREAD

#undef PTHREAD_DEBUG		/* just for debug */

/*** PTHREAD VERSION (experimental) ***/
/* main function for adin thread created at main.c */
/* see doc/thread.txt */
void
adin_cut_thread(ADIN_THREAD_P *arg)
{
  SP16 *buffer;
  int bp, bpmax;
  int len_limit, realtime_factor;
  int i, imax, len, stlen;
  int cnt;
  SP16 *speech;
  int *speechlen;
  boolean *transfer_flag;

  /* variables for zero-cross routines*/
  int zc;
  boolean is_valid_data = FALSE;
  int nc = 0;

  speech = arg->speech;
  speechlen = arg->len;
  transfer_flag = arg->tflag;
  
  /* there are 3 buffers:
   *   temporary storage queue: buffer[]
   *   cycle buffer for zero-cross counting: (in zc_e)
   *
   * Each samples are first read to buffer[], then passed to count_zc_e()
   * to find trigger.  Samples between trigger and end of speech are 
   * passwd to (*ad_process) with pointer to the first byte and its length.
   *
   * All procedure is done in pipe-line style.
   *
   */
  /* prepare temporary buffer (queue) */
  bpmax = MAXSPEECHLEN;
  buffer = (SP16 *)mymalloc(sizeof(SP16) * bpmax);
  bp = 0;

  /** initialize params. of zero cross method and etc **/
  init_count_zc_e(thres, c_length, c_offset);

  /* prepare audio device */
  if ((*ad_start)() == FALSE) {
    j_printerr("Error: failed to start recording\n");
    free(buffer);
    return;
  }

  realtime_factor = 1;		/* assume RTx 1 */
  len_limit = 16000;

  /* ---------------------------------------------- */
  /* main loop */
  for (;;) {

    /* Read samples as 16bit shorts (SP16) */
    /* bp: pointer to samples left in queue buffer */
    /* all samples in device are read at offset [bp] -> [0..len-1] */
    cnt = (*ad_read)(&(buffer[bp]), bpmax - bp);
    if (cnt < 0) {
      j_printerr("sample read error\n");
      exit(-1);
    }
    len = bp + cnt;
    /* j_printf("- read %d (%d-%d)\n",cnt,bp,len); */

    /* strip off invalid samples */
    if (arg->stflag) {
      stlen = strip_zero(&(buffer[bp]), cnt);
      if (stlen != cnt) {
	cnt = stlen;
	len = bp + cnt;
      }
    }

    /* adjust amounts of process per a loop */
    if (cnt > 16000) {		/* more than 1 sec. samples in device */
      /* queueing is getting delayed when process is slower than realtime. */
      /* now increases limits to the amount of processing samples per a loop */
      /* to avoid buffer overrun in AD-in device */
      realtime_factor++;
      len_limit = cnt / realtime_factor;
    }
    imax = (len < len_limit) ? len : len_limit;

    /* proceed for each 'wstep' steps */
    i = 0;
    while (i + wstep <= imax) {
      /* count zero-cross and flip switches */
      /* swap samples in buffer to that of (length) step ago */
      zc = count_zc_e(&(buffer[i]), wstep);
      if (zc > NOISE_ZEROCROSS) {
	nc = 0;                   /* reset nc counter */
	if (is_valid_data == FALSE) {
	  is_valid_data = TRUE;   /* start to record */
	}
      } else if (is_valid_data == TRUE) {
	/* noise after input */
	nc++;
	if (nc >= nc_max) {       /* end of loop */
	  /* not terminate, switch off transfer flag */
	  is_valid_data = FALSE;
	  *transfer_flag = FALSE;
#ifdef THREAD_DEBUG
	  printf("adin-mic: transfer turned off\n");
#endif
	}
      }
      if (is_valid_data == TRUE) {
	/* store data */
	if (*transfer_flag == TRUE) {
#ifdef THREAD_DEBUG
	  printf("adin-mic: write %d-%d\n", *speechlen, *speechlen + wstep);
#endif
	  if (*speechlen + wstep >= arg->maxlen) {
	    /* buffer overflow, stop now */
#ifdef THREAD_DEBUG
	    printf("adin-mic: transfer buffer overflow\n");
#endif
	    *transfer_flag = FALSE;
	  } else {
	    memcpy(&(speech[*speechlen]), &(buffer[i]), wstep * sizeof(SP16));
	    *speechlen += wstep;
	  }
	}
#ifdef THREAD_DEBUG
	else {
	  printf("adin-mic: valid but not transfered\n");
	}
#endif
      }
      i += wstep;
    }
    
    /* purge processed samples and update queue */
    if (i > 0) {
      memmove(buffer, &(buffer[i]), (len - i) * sizeof(SP16));
    }
    bp = len - i;
  }

  /* no termination: run forever */
}

#endif /* HAVE_PTHREAD */
