/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* rddata.c --- read waveform file */
/* sampling formats other than 16kHz,16bit,mono, will be rejected */

/* assume sizeof(int)=4, sizeof(short)=2 */

#include <sent/stddefs.h>
#include <sent/speech.h>


#ifdef HAVE_LIBSNDFILE
/* read in audio file using libsndfile interface */
/* From web site (http://www.zip.com.au/~erikd/libsndfile/):
     Libsndfile is a library designed to allow the reading and writing of
     many different sampled sound file formats (such as MS Windows WAV and
     the Apple/SGI AIFF format) through one standard library interface.

   version 0.0.17 has a capable of reading the following sound file
   formats:
   o Microsoft WAV 8, 16, 24 and 32 bit integer PCM. 
   o Microsoft 32 bit floating point PCM. 
   o Apple/SGI AIFF and AIFC uncompressed 8, 16, 24 and 32 bit integer PCM. 
   o Sun/NeXT AU/SND format (big endian 8, 16, 24 and 32 bit PCM,
     8 bit u-law and 8 bit A-law). 
   o Dec AU format (little endian 8, 16, 24 and 32 bit PCM, 8 bit u-law
     and 8 bit A-law). 
   o Microsoft IMA/DVI ADPCM WAV format (16 bits per sample compressed
     to 4 bits per sample). 
   o Microsoft ADPCM WAV format (16 bits per sample compressed to 4 bits
     per sample). 
   o Microsoft 8 bit A-law an u-law formats (16 bits per sample compressed
     to 8 bits per sample). 
   o RAW header-less PCM files of 8 (signed and unsigned), 16, 24 and 32
     bits. The 16, 24 and 32 bit
   o files may be big or little endian byte ordering. 
   o Amiga uncompressed IFF / 8SVX / 16SV PCM files (8 and 16 bit). 
   o Ensoniq PARIS big and little endian, 16 and 24 bit PCM files (.PAF).

   NOTES:
   1) Files are rejected if the sample formats are other than
      "16kHz,16bit,mono".
   2) Not all formats are tested.
   3) Licensing: the libsndfile library is distributed under GPL. */   

/* libsndfile header */
#include <sndfile.h>

/* check whether the format is 16kHz, 16bit, monoral */
static boolean
check_format(SF_INFO *s)
{
  if ((s->format & SF_FORMAT_TYPEMASK) != SF_FORMAT_RAW && s->samplerate != 16000) {
    j_printerr("rddata_sndfile: sample rate = %d Hz (!= 16000)\n", s->samplerate);
    return FALSE;
  }
  if (s->channels != 1) {
    j_printerr("rddata_sndfile: channel num = %d (!= 1 i.e. monoral)\n", s->channels);
    return FALSE;
  }
  if (s->pcmbitwidth != 16) {
    j_printerr("rddata_sndfile: %dbit sampling (!= 16bit)\n", s->pcmbitwidth);
    return FALSE;
  }
  return TRUE;
}  

/* rddata using libsndfile */
/* all endian conversion is done within the library */
int
rddata(char *fname, SP16 buf[], int *storelen, int maxlen)
{
  SF_INFO sinfo;		/* file info */
  SNDFILE *sp;

  /* for RAW file, initial header info must be filled by the caller */
  sinfo.channels = 1;
  sinfo.pcmbitwidth = 16;
  sinfo.samplerate = 16000;
  sinfo.format = SF_FORMAT_RAW | SF_FORMAT_PCM_BE; /* assume big endian */

  /* open and read header info */
  if ((sp = sf_open_read(fname, &sinfo)) == NULL) {
    sf_perror(sp);
    j_printerr("rddata: failed to open %s\n",fname);
    return(-1);
  }

  if (check_format(&sinfo) == FALSE) {
    j_printerr("rddata: invalid format in %s\n", fname);
    return(-1);
  }

  /* read in data */
  if ((*storelen = sf_read_short(sp, buf, sinfo.samples)) < 0) {
    sf_perror(sp);
    j_printerr("rddata: failed to read samples in %s\n",fname);
    return(-1);
  }

  if (sf_close(sp) != 0) {
    sf_perror(sp);
    j_printerr("rddata: failed to close %s\n",fname);
    return(-1);
  }

  return(0);
}  

#else  /* ~HAVE_LIBSNDFILE */

/************ original function that can read .raw and .wav file **********/

/* read data with endian conversion */
static boolean
myread(void *buf, size_t unitbyte, int unitnum, FILE *fp)
{
  int tmp;
  if ((tmp = myfread(buf, unitbyte, unitnum, fp)) < unitnum) {
    return(FALSE);
  }
#ifdef WORDS_BIGENDIAN
  /* all .wav datas are in little endian */
  swap_bytes(buf, unitbyte, unitnum);
#endif
  return(TRUE);
}

/* read in Microsoft WAVEfmt file */
static int
rddata_wav(FILE *fp, SP16 spbuf[], int *readlen, int maxlen)
{
  char dummy[9];
  unsigned int i, len;
  unsigned short s;
#define MYREAD(A,B,C,D)  if (!myread(A, B, C, D)) {j_printerr("rddata_wav: file is corrupted\n"); return -1;}

#if 0				/* already read in rddata() */
  /* first 4 byte: 'R' 'I' 'F' 'F' */
  MYREAD(dummy, 1, 4, fp);
  if (dummy[0] != 'R' ||
      dummy[1] != 'I' ||
      dummy[2] != 'F' ||
      dummy[3] != 'F') {
    j_printerr("rddata_wav: Not a 'RIFF' format\n");
    fclose_readfile(fp); return -1;
  }
#endif
  /* 4 byte: byte num of rest ( = filesize - 8) */
  /* --- just skip */
  MYREAD(dummy, 1, 4, fp);
  /* first part: WAVE format specifications */
  /* 8 byte: 'W' 'A' 'V' 'E' 'f' 'm' 't' ' ' */
  MYREAD(dummy, 1, 8, fp);
  if (dummy[0] != 'W' ||
      dummy[1] != 'A' ||
      dummy[2] != 'V' ||
      dummy[3] != 'E' ||
      dummy[4] != 'f' ||
      dummy[5] != 'm' ||
      dummy[6] != 't' ||
      dummy[7] != ' ') {
    j_printerr("rddata_wav: WAVEfmt part not found, file corrupted?\n");
    fclose_readfile(fp); return -1;
  }
  /* 4byte: byte size of this part (fixed to 16) */
  MYREAD(&i, 4, 1, fp);
  if (i != 16) {
    j_printerr("rddata_wav: unknown WAVEfmt part, length other than 16 bytes\n");
    fclose_readfile(fp); return -1;
  }
  /* 2byte: data format */
  MYREAD(&s, 2, 1, fp);
  if (s != 1) {
    j_printerr("rddata_wav: data format != PCM (id=%d)\n", s);
    fclose_readfile(fp); return -1;
  }
  /* 2byte: channel num */
  MYREAD(&s, 2, 1, fp);
  if (s >= 2) {
    j_printerr("rddata_wav: channel num != 1 (%d)\n", s);
    fclose_readfile(fp); return -1;
  }
  /* 4byte: sampling rate */
  MYREAD(&i, 4, 1, fp);
  if (i != 16000) {
    j_printerr("rddata_wav: sampling rate != 16000 (%d)\n", i);
    fclose_readfile(fp); return -1;
  }
  /* 4byte: bytes per second */
  MYREAD(&i, 4, 1, fp);
  if (i != 32000) {
    j_printerr("rddata_wav: bytes per second != 32000 (%d)\n", i);
    fclose_readfile(fp); return -1;
  }
  /* 2bytes: bytes per frame ( = (bytes per sample) x channel ) */
  MYREAD(&s, 2, 1, fp);
  if (s != 2) {
    j_printerr("rddata_wav: (bytes per sample) x channel != 2 (%d)\n", s);
    fclose_readfile(fp); return -1;
  }
  /* 2bytes: bits per sample */
  MYREAD(&s, 2, 1, fp);
  if (s != 16) {
    j_printerr("rddata_wav: bits per sample != 16 (%d)\n", s);
    fclose_readfile(fp); return -1;
  }
  /* end of WAVEfmt part */

  /* seek for 'data' part */
  while (myread(&dummy, 1, 4, fp)) {
    MYREAD(&len, 4, 1, fp);
    if (dummy[0] == 'd' &&
	dummy[1] == 'a' &&
	dummy[2] == 't' &&
	dummy[3] == 'a') {
      break;
    }
    for (i=0;i<len;i++) myread(&dummy, 1, 1, fp);
  }
  /* read in "data" part --- this is speech data */
  len /= sizeof(SP16);
  if (len > maxlen) {
    j_printerr("rddata_wav: too long (%d samples > %d)\n", len, maxlen);
    fclose_readfile(fp); return -1;
  }
  if (!myread(spbuf, sizeof(SP16), len, fp)) {
    j_printerr("rddata_wav: data part corrupted\n");
    fclose_readfile(fp); return -1;
  }
  *readlen = len;

  return(0);
}

/* Read sample data from file 'fname' and store to 'a[]' */
/* 'maxlen' is the maximum sample num of 'a[]' (not bytes, but sample num!) */
/* When the file is longer than 'maxlen', the rests are discarded */
/* Actual read sample num will be stored in '*len' */
/* Returns 0 on success, -1 on failure */
int
rddata(char  *fname, SP16 data[], int   *len, int   maxlen)
{
  FILE *fp;
  char dummy[4];
  int adv;
  
  if ((fp = fopen_readfile(fname)) == NULL) {
    perror("rddata");
    j_printerr("failed to open %s\n",fname);
    return(-1);
  }
  /* check first 4 byte to detect Microsoft WAVE format */
  if (myfread(dummy, 1, 4, fp) < 4) {
    j_printerr("Error: size less than 4 bytes?\n",fname);
    fclose_readfile(fp);
    return(-1);
  }
  if (dummy[0] == 'R' &&
      dummy[1] == 'I' &&
      dummy[2] == 'F' &&
      dummy[3] == 'F') {
    /* read as WAVE file */
    if (rddata_wav(fp, data, len, maxlen) != 0) {
      j_printerr("Error: failed to read %s\n",fname);
      fclose_readfile(fp);
      return(-1);
    }
  } else {
    /* read as raw format file */
    memcpy(data, dummy, 4);    /* already read (4/sizeof(SP)) samples */
    adv = 4 / sizeof(SP16);
    *len = myfread(&(data[adv]), sizeof(SP16), maxlen - adv, fp) + adv;
    if (!feof(fp)) {
      j_printerr("Error: too long input ( > %d samples)\n", maxlen);
      fclose_readfile(fp);
      return(-1);
    }
#ifndef WORDS_BIGENDIAN		/* LITTLE */
    /* assume input is BIG ENDIAN */
    swap_sample_bytes(data, *len);
#endif
  }
  /* j_printerr("%d sample (%d bytes)\n", *len, (*len) * sizeof(SP16));	*/
  fclose_readfile(fp);
  return(0);
}


#endif  /* ~HAVE_LIBSNDFILE */
