/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* rdhmmdef.c --- read in HTK format HMM definition file (main) */

/* $Id: rdhmmdef.c,v 1.2 1998/09/02 15:47:20 ri Exp $ */

#include <sent/stddefs.h>
#include <sent/htk_param.h>
#include <sent/htk_hmm.h>

#define MAXBUFLEN  4096
#ifdef DELM
#undef DELM
#endif
#define DELM " \t\n<>"

char *rdhmmdef_token;		/* token */
				/* referred by other rdhmmdef_*.c */

static char *buf = NULL;
static int line;

void
rderr(char *str)
{
  if (rdhmmdef_token == NULL) {	/* end of file */
    j_printerr("\nError: %s on end of file\n", str);
  } else {
    j_printerr("\nError at line %d: %s\n", line, (str) ? str : "parse error");
  }
  exit(1);
}

/* read next token and set it to rdhmmdef_token */
/* return NULL on end of file */
char *
read_token(FILE *fp)
{
  /* the quotation handling is moved to
     mystrtok_quote() in util/mystrtok.c */
  
  if (buf != NULL) {
    /* already have buffer */
    if ((rdhmmdef_token = mystrtok_quote(NULL)) != NULL) {
      /* return next token */
      return rdhmmdef_token;
    }
  } else {
    /* init: allocate buffer for the first time */
    buf = (char *)mymalloc(MAXBUFLEN);
    line = 1;
  }
  /* read new 1 line */
  if (getl(buf, MAXBUFLEN, fp) == NULL) {
    rdhmmdef_token = NULL;
  } else {
    rdhmmdef_token = mystrtok_quote(buf);
    line++;
  }
  return rdhmmdef_token;
}


#ifdef LOGTABLE
/* $BA+0\3NN($rBP?t2=(B */
static void
conv_log_arc(HTK_HMM_INFO *hmm)
{
  HTK_HMM_Trans *tr;
  int i,j;
  LOGPROB l;

  for (tr = hmm->trstart; tr; tr = tr->next) {
    for(i=0;i<tr->statenum;i++) {
      for(j=0;j<tr->statenum;j++) {
	l = tr->a[i][j];
	tr->a[i][j] = (l != 0.0) ? log10(l) : LOG_ZERO;
      }
    }
  }
}
#endif


/* init */
void
init_hmm(HTK_HMM_INFO *hmm)
{
  /* set default options */
  hmm->opt.stream_info.num = 1;
  hmm->opt.cov_type = C_DIAG_C;
  hmm->opt.dur_type = D_NULL;

  hmm->trstart = NULL;
  hmm->vrstart = NULL;
  hmm->ststart = NULL;
  hmm->dnstart = NULL;
  hmm->start   = NULL;
  hmm->lgstart = NULL;
  hmm->physical_root = NULL;
  hmm->logical_root = NULL;
  hmm->tr_root = NULL;
  hmm->vr_root = NULL;
  hmm->dn_root = NULL;
  hmm->st_root = NULL;
  hmm->codebooknum = 0;
  hmm->codebook_root = NULL;
  hmm->maxcodebooksize = 0;
  hmm->totalmixnum = 0;
  hmm->totalstatenum = 0;
  hmm->totalhmmnum = 0;
  hmm->totallogicalnum = 0;
  hmm->is_triphone = FALSE;
  hmm->is_tied_mixture = FALSE;
}

/* main routine */
boolean
rdhmmdef(FILE *fp, HTK_HMM_INFO *hmm)
{
  char macrosw;
  char *name;

  /* HMM$B9=B$BN$N=i4|2=(B */
  init_hmm(hmm);
  
  /* $B:G=i$N(B token $BFI$_9~$_(B */
  read_token(fp);
  
  /* $B%H%C%W%l%Y%k$O$9$Y$F(B "~" $B$G;O$^$kDj5A(B */
  while (rdhmmdef_token != NULL) {/* break on EOF */
    if (rdhmmdef_token[0] != '~') {
      rderr(NULL);
    }
    macrosw = rdhmmdef_token[1];
    read_token(fp);		/* "~.." $B$N<!$N%H!<%/%s$rFI$s$G$*$/(B */
    switch(macrosw) {
    case 'o':			/* global option */
      set_global_opt(fp,hmm);
      break;
    case 't':			/* transition macro */
      name = mybstrdup(rdhmmdef_token);
      read_token(fp);
      def_trans_macro(name, fp, hmm);
      break;
    case 's':			/* state macro */
      name = mybstrdup(rdhmmdef_token);
      read_token(fp);
      def_state_macro(name, fp, hmm);
      break;
    case 'm':			/* density (mixture) macro */
      name = mybstrdup(rdhmmdef_token);
      read_token(fp);
      def_dens_macro(name, fp, hmm);
      break;
    case 'h':			/* HMM define */
      name = mybstrdup(rdhmmdef_token);
      read_token(fp);
      def_HMM(name, fp, hmm);
      break;
    case 'v':			/* Variance macro */
      name = mybstrdup(rdhmmdef_token);
      read_token(fp);
      def_var_macro(name, fp, hmm);
    }
  }

  if (check_all_hmm_limit(hmm)) {
    /* passed */
    j_printerr("limit check passed...\n");
  } else {
    j_printerr("Error: cannot use this HMM for system limitation.\n");
    exit(1);
  }
#ifdef LOGTABLE
  /* $BA+0\3NN($rBP?t2=(B */
  conv_log_arc(hmm);
#endif
  
  /* check if this system can handle this type of duration/covmatrix type */
  if (!check_hmm_options(hmm)) {
    exit(1);
  }

  /* $BA4(BHTK_HMM_State $B$KDL$7HV9f(B(id)$B$r$D$1$k(B */
  {
    HTK_HMM_State *stmp;
    int n;
    n = 0;
    for (stmp = hmm->ststart; stmp; stmp = stmp->next) {
      stmp->id = n++;
      if (n >= MAX_STATE_NUM) {
	j_printerr("Error: too much states > %d\n", MAX_STATE_NUM);
	exit(1);
      }
    }
    hmm->totalstatenum = n;
  }
  /* HMM $BAm?t$r7W;;(B */
  {
    HTK_HMM_Data *dtmp;
    int n;
    n = 0;
    for (dtmp = hmm->start; dtmp; dtmp = dtmp->next) n++;
    hmm->totalhmmnum = n;
  }
  /* mixture$BAm?t$r?t$($k(B */
  {
    HTK_HMM_Dens *dtmp;
    int n = 0;
    for (dtmp = hmm->dnstart; dtmp; dtmp = dtmp->next) {
      n++;
    }
    hmm->totalmixnum = n;
  }

  return(TRUE);
}
