/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* outprob_tm.c --- calculate outprob for tied-mixture model */

/* gaussian pruning algorithm: per-gaussian, threshold by heuristics */

/*
 * 1) Compute outprob for all Gaussian densities with pruning.
 *    (several mixtures (=codebooks) can have their own threshold)
 * 2) Store best scores in mixture-level cache, with score and density-id.
 * 3) HMM states ask the cache for their component densities and calc the
 *    output probs.
 *
 * cache are stored for each frame, so that it can be accessed at 2nd pass.
 *
 */

/* enable/disable experimental features */
/* THRES_AT_1 deleted */
/* LAST_BEST included */
#undef DIAG			/* output cache status */
#undef DIAG2			/* more detailed */

#include <sent/stddefs.h>
#include <sent/htk_hmm.h>
#include <sent/htk_param.h>
#include <sent/hmm.h>
#include <sent/gprune.h>

/* cache for each frame in outprob.c */
extern MIXCACHE ***mixture_cache; /* [bookid][time][#] (sequencial) */
extern int mix_n;		/* num of density to store in a frame */
#ifdef USE_ADDLOG_ARRAY
extern LOGPROB *probs;
#endif

static int booknum;
static LOGPROB *backmax; /* maximum backward sum for each dimension (reversed) */
static int backmax_num;
#ifdef DIAG2
static int total_cnum;
static int calced_cnum;
#endif

/* initialize cache area (should be called once on startup) */
void
mixture_cache_init_heuristic(HTK_HMM_INFO *hmminfo, int best_mixture_num)
{
  mix_n = best_mixture_num;
  booknum = hmminfo->codebooknum;

  backmax_num = hmminfo->opt.vec_size + 1;
  backmax = (LOGPROB *)mymalloc(sizeof(LOGPROB) * backmax_num);
    
#ifdef USE_ADDLOG_ARRAY
  probs = (LOGPROB *)mymalloc(sizeof(LOGPROB) * mix_n);
#endif
}

/* prepare cache area (should be called for each sample) */
void
mixture_cache_prepare_heuristic(int framenum)
{
  static int oldframenum;
  int bid, t, size;

  if (mixture_cache != NULL) {
    for(t=0;t<oldframenum;t++) {
      free(mixture_cache[t][0]);
      free(mixture_cache[t]);
    }
    free(mixture_cache);
  }
  size = mix_n * booknum;
  
  mixture_cache = (MIXCACHE ***)mymalloc(sizeof(MIXCACHE **) * framenum);
  for(t=0;t<framenum;t++) {
    mixture_cache[t] = (MIXCACHE **)mymalloc(sizeof(MIXCACHE *) * booknum);
    mixture_cache[t][0] = (MIXCACHE *)mymalloc(sizeof(MIXCACHE) * size);
    for(bid=1;bid<booknum;bid++) {
      mixture_cache[t][bid] = &(mixture_cache[t][0][mix_n * bid]);
    }
  }

  oldframenum = framenum;
}


/* general cache operations */
static int
find_insert_point(MIXCACHE *c, LOGPROB score, int len)
{
  /* binary search on score */
  int left = 0;
  int right = len - 1;
  int mid;

  while (left < right) {
    mid = (left + right) / 2;
    if (c[mid].score > score) {
      left = mid + 1;
    } else {
      right = mid;
    }
  }
  return(left);
}

static int
cache_push(MIXCACHE *cache, int id, LOGPROB score, int len)
{
  int insertp;

  if (len == 0) {		/* first one */
    cache[0].score = score;
    cache[0].id = id;
    return(1);
  }
  if (cache[len-1].score >= score) { /* bottom */
    if (len < mix_n) {		/* append to bottom */
      cache[len].score = score;
      cache[len].id = id;
      len++;
    }
    return len;
  }
  if (cache[0].score < score) {
    insertp = 0;
  } else {
    insertp = find_insert_point(cache, score, len);
  }
  if (len < mix_n) {
    memmove(&(cache[insertp+1]), &(cache[insertp]), sizeof(MIXCACHE)*(len - insertp));
  } else if (insertp < len - 1) {
    memmove(&(cache[insertp+1]), &(cache[insertp]), sizeof(MIXCACHE)*(len - insertp - 1));
  }
  cache[insertp].score = score;
  cache[insertp].id = id;
  if (len < mix_n) len++;
  return(len);
}


/* mail routines to calculate mixtures for given frame, and store to cache */
/*

  best_mixtures_on_last_frame[]
  
  dim:  0 1 2 3 4 .... veclen-1    $B"*(B sum up
 ================================
  backsum $B"+"+"+"+"+"+"+"+"+(B
 --------------------------------
  mix1  $B"*"*(B
  mix2  $B"*"*(B
  mix3  $B"*"*(B
  ...
  mixN  $B"*"*(B
 ================================
  score = forward sum + backmax

  algorithm 1:

     // store sub-optimal max prob for each dimension
     init backsum[]
     foreach all_mixtures in best_mixtures_on_last_frame {
        foreach dim {
           p = prob of this dim
	   if (backmax[dim] < p) backmax[dim] = p
	   sumup score
        }
	store to cache
     }
     // sum-up backward to make backmax
     foreach dim-1 backward {
        backmax[dim] += backmax[dim+1]
     }
     // compute rest with pruning with heuristics
     threshold = the current lowest score
     foreach rest_mixtures {
        foreach dim {
	   sumup score
	   if (score + backmax[dim+1] < threshold) {
	      skip this mixture
	   }
	}
	update threshold
     }
  }
     
*/

/* calc_contprob with Gaussing threshold --- for first frame */
static LOGPROB
calc_contprob_with_pruning(HTK_HMM_Dens *binfo, VECT *vec, short veclen, LOGPROB thres)
{
  VECT tmp, x;
  VECT *mean;
  VECT *var;
  VECT fthres = thres * (-2.0);
#ifdef DIAG2
  int totallen = veclen;
#endif

  if (binfo == NULL) return(LOG_ZERO);
  mean = binfo->mean;
  var = binfo->var->vec;
#ifdef DIAG2
  total_cnum += veclen;
#endif

  tmp = binfo->gconst;
  for (; veclen > 0; veclen--) {
    x = *(vec++) - *(mean++);
    tmp += x * x / *(var++);
    if ( tmp > fthres) {
#ifdef DIAG2
      calced_cnum += totallen - veclen + 1;
#endif
      return LOG_ZERO;
    }
  }
#ifdef DIAG2
  calced_cnum += totallen;
#endif
  return(tmp / -2.0);
}

/* initialize backmax */
static void
init_backmax(int len)
{
  int i;
  for(i=0;i<len;i++) backmax[i] = -(LOG_ZERO); /* negatived */
}
/* sum-up backward to make backmax */
/*                        |
 *  0 1 2 ... len-2 len-1 | len
 *
 *  a b c      x      y   | ????
 *            $B"-(B
 *  .....     x+y     y     0.0
 */
static void
make_backmax(int len)
{
  int i;
  backmax[len] = 0.0;
  /* backmax[len-1] = backmax[len-1];*/
  for(i=len-2;i>=0;i--) {
    backmax[i] += backmax[i+1];
  }
  /*  for(i=0;i<=len;i++) {
    printf("backmax[%d]=%f\n",i,backmax[i]);
    }*/
}

/* calculate prob with setting maxprob for each dimension */
static LOGPROB
calc_contprob_with_dimmax(HTK_HMM_Dens *binfo, VECT *vec, short veclen)
{
  LOGPROB sum = 0.0, tmp, x;
  VECT *mean;
  VECT *var;
  LOGPROB *bm = backmax;

  if (binfo == NULL) return(LOG_ZERO);
  mean = binfo->mean;
  var = binfo->var->vec;
#ifdef DIAG2
  total_cnum += veclen;
  calced_cnum += veclen;
#endif

  for (; veclen > 0; veclen--) {
    x = *(vec++) - *(mean++);
    tmp = x * x / *(var++);
    sum += tmp;
    if ( *bm > tmp) *bm = tmp;
    bm++;
  }
  return((sum + binfo->gconst) / -2.0);
}

/* calculate prob with pruning with heuristics */
static LOGPROB
calc_contprob_with_pruning_backmax(HTK_HMM_Dens *binfo, VECT *vec, short veclen, LOGPROB thres)
{
  LOGPROB tmp, x;
  VECT *mean;
  VECT *var;
  LOGPROB *bm = backmax;
  LOGPROB fthres;

  if (binfo == NULL) return(LOG_ZERO);
  mean = binfo->mean;
  var = binfo->var->vec;
  fthres = thres * (-2.0) - binfo->gconst;
#ifdef DIAG2
  int totallen = veclen;
#endif

#ifdef DIAG2
  total_cnum += veclen;
#endif

  tmp = 0.0;
  bm++;
  for (; veclen > 0; veclen--) {
    x = *(vec++) - *(mean++);
    tmp += x * x / *(var++);
    if ( tmp + *bm > fthres) {
#ifdef DIAG2
      calced_cnum += totallen - veclen + 1;
#endif
      return LOG_ZERO;
    }
    bm++;
  }
#ifdef DIAG2
  calced_cnum += totallen;
#endif
  return((tmp + binfo->gconst) / -2.0);
}


static MIXCACHE **last_tcache;	/* pointer to mixture cache of last frame */
static boolean *mixcalced = NULL; /* [mid] = TRUE if calced */
static int maxbooksize;		/* size of above */
/* input data access abbreviations for speed */
/* set in calc_all_mixtures_and_cache(), used in callback_calc_codebook_and_cache */
static VECT *tvec;
static MIXCACHE **tcache;
static int tveclen;

/* calc scores for given codebook with pruning by number */
/* and store to cache */
static void
callback_calc_codebook_and_cache(void *bookp)
{
  GCODEBOOK *book = bookp;
  LOGPROB score, thres;
  int i, num;			/* num = cached num */
  int j;

#ifdef DIAG
  printf("book \"%s\"(%d)\n", book->name, book->num);
#endif

  num = 0;
  /* prepare buffer to mark densities whether calculated or not in this frame */
  if (mixcalced == NULL) {	/* allocate new (once in a run) */
    maxbooksize = book->num;
    mixcalced = (boolean *)mymalloc(sizeof(boolean)*maxbooksize);
    for(i=0;i<maxbooksize;i++) mixcalced[i] = FALSE;
  } else if (book->num > maxbooksize) {	/* expand if needed */
    maxbooksize = book->num;
    mixcalced = (boolean *)myrealloc(mixcalced, sizeof(boolean)*maxbooksize);
    for(i=0;i<maxbooksize;i++) mixcalced[i] = FALSE;
  }
  
  if (last_tcache != NULL) {	/* not the first frame */
    /* 1. calculate those within cache of last frame  */
    /*    and set maximum prob for each dimension */
    init_backmax(tveclen);
    for (j=0; j<mix_n; j++) {
      i = last_tcache[book->id][j].id;
      /* calculate and set thres for each dimension */
      score = calc_contprob_with_dimmax(book->d[i], tvec, tveclen);
      num = cache_push(tcache[book->id], i, score, num);
      mixcalced[i] = TRUE;	/* mark as calculated */
    }
    /* 2. make backmax by summing-up probs backwards */
    make_backmax(tveclen);
    /* 3. calculate the rest with pruning with backmax heuristic */
    thres = tcache[book->id][num-1].score;
    for (i = 0; i<book->num; i++) {
      if (mixcalced[i]) {
	mixcalced[i] = FALSE;	/* clear for next frame */
	continue;
      }
      score = calc_contprob_with_pruning_backmax(book->d[i], tvec, tveclen, thres);
      if (score > LOG_ZERO) {
      /* update cache and threshold */
	num = cache_push(tcache[book->id], i, score, num);
	thres = tcache[book->id][num-1].score;
      }
    }
  } else {
    /* for the first frame, normal algorthm is used below */
    /* calculate each density with pruning by score */
    for (i=0; i<book->num; i++) {
      if (num >= mix_n) {
	score = calc_contprob_with_pruning(book->d[i], tvec, tveclen, thres);
	if (score < thres) continue;
      } else {
	score = calc_contprob(book->d[i], tvec, tveclen);
#ifdef DIAG2
	total_cnum += tveclen;
	calced_cnum += tveclen;
#endif
      }
      /* update cache and threshold */
      num = cache_push(tcache[book->id], i, score, num);
      if (num >= mix_n) {
	thres = tcache[book->id][num-1].score;
      }
    }
  }

#ifdef DIAG
  {
    int j;
    printf("cache:");
    for(j=0;j<num;j++) {
      printf(" %f(%d)", tcache[book->id][j].score, tcache[book->id][j].id);
    }
    printf("\n");
  }
#endif

  if (num < mix_n) {
    j_printerr("Error: cached mixture num < thres num (%d)\n", mix_n);
    exit(1);
  }

}

/* calc Gaussian density outprobs for all mixtures (= codebook) */
void
calc_all_mixtures_and_cache_heuristic(int t, HTK_Param *param, HTK_HMM_INFO *hmminfo)
{
#ifdef DIAG
  printf("### %d th\n", t);
#endif
#ifdef DIAG2
  total_cnum = calced_cnum = 0;
#endif
  /* store vector and cache info to temporary pointers */
  tvec = param->parvec[t];
  tveclen = param->veclen;
  tcache = mixture_cache[t];
  last_tcache = (t > 0) ? mixture_cache[t-1] : NULL;
  /* call main calculation functions for each codebook */
  aptree_traverse_and_do(hmminfo->codebook_root, callback_calc_codebook_and_cache);
#ifdef DIAG2
  printf("calced %6d/%6d (%d%%)\n", calced_cnum, total_cnum,
	 calced_cnum * 100 / total_cnum);
#endif

}
