/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* outprob_tm.c --- calculate outprob for tied-mixture model */

/* gaussian pruning algorithm: per-gaussian, threshold by current N-best gaussian */
/* variants: initial threshold by last_best_gaussians */

/*
 * 1) Compute outprob for all Gaussian densities with pruning.
 *    (several mixtures (=codebooks) can have their own threshold)
 * 2) Store best scores in mixture-level cache, with score and density-id.
 * 3) HMM states ask the cache for their component densities and calc the
 *    output probs.
 *
 * cache are stored for each frame, so that it can be accessed at 2nd pass.
 *
 */

/* enable/disable experimental features */
#undef THRES_AT_1		/* pruning on only 1 vector ... bad */
#define LAST_BEST		/* compute bests of last frame first */
				/* to determine thresholds */
#undef DIAG			/* output cache status */
#undef DIAG2			/* more detailed */

#include <sent/stddefs.h>
#include <sent/htk_hmm.h>
#include <sent/htk_param.h>
#include <sent/hmm.h>
#include <sent/gprune.h>

/* cache for each frame in outprob.c */
extern MIXCACHE ***mixture_cache; /* [bookid][time][#] (sequencial) */
extern int mix_n;		/* num of density to store in a frame */
#ifdef USE_ADDLOG_ARRAY
extern LOGPROB *probs;
#endif

static int booknum;
#ifdef DIAG2
static int total_cnum;
static int calced_cnum;
#endif



/* calc_contprob with Gaussing threshold */
static LOGPROB
calc_contprob_with_pruning(HTK_HMM_Dens *binfo, float *vec, short veclen, LOGPROB thres)
{
  float tmp, x;
  float *mean;
  float *var;
  float fthres = thres * (-2.0);
#ifdef DIAG2
  int totallen = veclen;
#endif

  if (binfo == NULL) return(LOG_ZERO);
  mean = binfo->mean;
  var = binfo->var->vec;

#ifdef DIAG2
  total_cnum += veclen;
#endif

  tmp = binfo->gconst;
  for (; veclen > 0; veclen--) {
    x = *(vec++) - *(mean++);
    tmp += x * x / *(var++);
    if ( tmp > fthres) {
#ifdef DIAG2
      calced_cnum += totallen - veclen + 1;
#endif
      return LOG_ZERO;
    }
  }
#ifdef DIAG2
  calced_cnum += totallen;
#endif
  return(tmp / -2.0);
}


/* initialize cache area (should be called once on startup) */
void
mixture_cache_init_safe(HTK_HMM_INFO *hmminfo, int best_mixture_num)
{
  mix_n = best_mixture_num;
  booknum = hmminfo->codebooknum;

#ifdef USE_ADDLOG_ARRAY
  probs = (LOGPROB *)mymalloc(sizeof(LOGPROB) * mix_n);
#endif
}

/* prepare cache area (should be called for each sample) */
void
mixture_cache_prepare_safe(int framenum)
{
  static int oldframenum;
  int bid, t, size;

  if (mixture_cache != NULL) {
    for(t=0;t<oldframenum;t++) {
      free(mixture_cache[t][0]);
      free(mixture_cache[t]);
    }
    free(mixture_cache);
  }
  size = mix_n * booknum;
  
  mixture_cache = (MIXCACHE ***)mymalloc(sizeof(MIXCACHE **) * framenum);
  for(t=0;t<framenum;t++) {
    mixture_cache[t] = (MIXCACHE **)mymalloc(sizeof(MIXCACHE *) * booknum);
    mixture_cache[t][0] = (MIXCACHE *)mymalloc(sizeof(MIXCACHE) * size);
    for(bid=1;bid<booknum;bid++) {
      mixture_cache[t][bid] = &(mixture_cache[t][0][mix_n * bid]);
    }
  }

  oldframenum = framenum;
}


/* general cache operations */
static int
find_insert_point(MIXCACHE *c, LOGPROB score, int len)
{
  /* binary search on score */
  int left = 0;
  int right = len - 1;
  int mid;

  while (left < right) {
    mid = (left + right) / 2;
    if (c[mid].score > score) {
      left = mid + 1;
    } else {
      right = mid;
    }
  }
  return(left);
}

static int
cache_push(MIXCACHE *cache, int id, LOGPROB score, int len)
{
  int insertp;
  int i;

  if (len == 0) {		/* first one */
    cache[0].score = score;
    cache[0].id = id;
    return(1);
  }
  if (cache[len-1].score >= score) { /* bottom */
    if (len < mix_n) {		/* append to bottom */
      cache[len].score = score;
      cache[len].id = id;
      len++;
    }
    return len;
  }
  if (cache[0].score < score) {
    insertp = 0;
  } else {
    insertp = find_insert_point(cache, score, len);
  }
  if (len < mix_n) {
    bcopy(&(cache[insertp]), &(cache[insertp+1]), sizeof(MIXCACHE)*(len - insertp));
  } else if (insertp < len - 1) {
    bcopy(&(cache[insertp]), &(cache[insertp+1]), sizeof(MIXCACHE)*(len - insertp - 1));
  }
  cache[insertp].score = score;
  cache[insertp].id = id;
  if (len < mix_n) len++;
  return(len);
}


/* mail routines to calculate mixtures for given frame, and store to cache */

#ifdef LAST_BEST
static MIXCACHE **last_tcache;	/* pointer to mixture cache of last frame */
static boolean *mixcalced = NULL; /* [mid] = TRUE if calced */
static int maxbooksize;		/* size of above */
#endif
/* input data access abbreviations for speed */
/* set in calc_all_mixtures_and_cache(), used in callback_calc_codebook_and_cache */
static float *tvec;
static MIXCACHE **tcache;
static int tveclen;

/* calc scores for given codebook with pruning by number */
/* and store to cache */
static void
callback_calc_codebook_and_cache(void *bookp)
{
  GCODEBOOK *book = bookp;
  LOGPROB score, thres;
  int i, num;			/* num = cached num */
#ifdef LAST_BEST
  int j;
#endif

#ifdef DIAG
  printf("book \"%s\"(%d)\n", book->name, book->num);
#endif

  num = 0;
  
#ifdef LAST_BEST
  /* compute bests of last frame first */

  /* prepare buffer to mark densities whether calculated or not in this frame */
  if (mixcalced == NULL) {	/* allocate new (once in a run) */
    maxbooksize = book->num;
    mixcalced = (boolean *)mymalloc(sizeof(boolean)*maxbooksize);
    for(i=0;i<maxbooksize;i++) mixcalced[i] = FALSE;
  } else if (book->num > maxbooksize) {	/* expand if needed */
    maxbooksize = book->num;
    mixcalced = (boolean *)myrealloc(mixcalced, sizeof(boolean)*maxbooksize);
    for(i=0;i<maxbooksize;i++) mixcalced[i] = FALSE;
  }
  
  if (last_tcache != NULL) {	/* not the first frame */
    /* 1. calculate those within cache of last frame  */
    /*    and set initial threshold */
    for (j=0; j<mix_n; j++) {
      i = last_tcache[book->id][j].id;
      score = calc_contprob(book->d[i], tvec, tveclen);
#ifdef DIAG2
      total_cnum += tveclen;
      calced_cnum += tveclen;
#endif
      num = cache_push(tcache[book->id], i, score, num);
      mixcalced[i] = TRUE;	/* mark them as calculated */
    }
    thres = tcache[book->id][num-1].score;
    /* 2. calculate the rest */
    for (i = 0; i<book->num; i++) {
      if (mixcalced[i]) {
	mixcalced[i] = FALSE;	/* to keep clear when done */
	continue;
      }
#ifdef THRES_AT_1
      /* pruning with only 1-d vector */
      score = calc_contprob_with_pruning(book->d[i], tvec, 1, thres);
#else
      score = calc_contprob_with_pruning(book->d[i], tvec, tveclen, thres);
#endif
      if (score <= thres) continue;
      /* update cache and threshold */
      num = cache_push(tcache[book->id], i, score, num);
      thres = tcache[book->id][num-1].score;
    }
  } else {
    /* for the first frame, normal algorthm is used below */

#endif /* LAST_BEST */

    /* calculate each density with pruning by score */
    for (i=0; i<book->num; i++) {
      if (num >= mix_n) {
#ifdef THRES_AT_1
      /* pruning with only 1-d vector */
	score = calc_contprob_with_pruning(book->d[i], tvec, 1, thres);
#else
	score = calc_contprob_with_pruning(book->d[i], tvec, tveclen, thres);
#endif
	if (score < thres) continue;
      } else {
	score = calc_contprob(book->d[i], tvec, tveclen);
#ifdef DIAG2
	total_cnum += tveclen;
	calced_cnum += tveclen;
#endif
      }
      /* update cache and threshold */
      num = cache_push(tcache[book->id], i, score, num);
      if (num >= mix_n) {
	thres = tcache[book->id][num-1].score;
      }
    }
#ifdef LAST_BEST
  }
#endif

#ifdef DIAG
  {
    int j;
    printf("cache:");
    for(j=0;j<num;j++) {
      printf(" %f(%d)", tcache[book->id][j].score, tcache[book->id][j].id);
    }
    printf("\n");
  }
#endif

  if (num < mix_n) {
    j_printerr("Error: cached mixture num < thres num (%d)\n", mix_n);
    exit(1);
  }

#ifdef THRES_AT_1
  /* calculate the rest for survived densities */
  {
    int j;
    for(j=0;j<mix_n;j++) {
      tcache[book->id][j].score = calc_contprob(book->d[tcache[book->id][j].id], tvec, tveclen);
    }
  }
#endif
  
}

/* calc Gaussian density outprobs for all mixtures (= codebook) */
void
calc_all_mixtures_and_cache_safe(int t, HTK_Param *param, HTK_HMM_INFO *hmminfo)
{
#ifdef DIAG
  printf("### %d th\n", t);
#endif
#ifdef DIAG2
  total_cnum = calced_cnum = 0;
#endif
  /* store vector and cache info to temporary pointers */
  tvec = param->parvec[t];
  tveclen = param->veclen;
  tcache = mixture_cache[t];
#ifdef LAST_BEST
  last_tcache = (t > 0) ? mixture_cache[t-1] : NULL;
#endif
  /* call main calculation functions for each codebook */
  aptree_traverse_and_do(hmminfo->codebook_root, callback_calc_codebook_and_cache);
#ifdef DIAG2
  printf("calced %6d/%6d (%d%%)\n", calced_cnum, total_cnum,
	 calced_cnum * 100 / total_cnum);
#endif
}
